"""
Utils for stat analysis
"""

from typing import Literal, List, Callable
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy.stats import shapiro

from abtoolkit.discrete.utils import estimate_ci_binomial


def check_clt(
    variable,
    do_plot_distribution: bool = True,
    sample_size: int = None,
    experiments_num: int = 5000,
    metric_f: Callable[[List, any], float] = np.mean,
    **metric_f_kwargs,
) -> float:
    """
    Tests the central limit theorem by sampling subsamples of length `sample_size`, `experiments_num` times.
    `metric_f` is then taken from a subsample and the distribution of such values is compared to the normal
    distribution using the Shapiro-Wilk test. Returns p-value, where the null hypothesis that the weights were drawn
    from a normal distribution and alternative is another distribution. So if p-value lower 0.05 then variable doesn't
    consider CLT.
    :param variable: array-like, variable we want to test
    :param sample_size: subsample size we take on each simulation step, if None then take length(variable).
    Default = None
    :param experiments_num: number of simulations for estimation, default = 5000
    :param metric_f: function to take from subsample, default = numpy.mean(). Additional args could be given using
    metric_kwargs parameter
    :param do_plot_distribution: whether to plot distribution or not, default = True
    :return: p-value of Shapiro-Wilk test for normality
    """

    sample_size = len(variable) if sample_size is None else sample_size
    values = []
    for _ in tqdm(range(experiments_num)):
        sample = np.random.choice(variable, size=sample_size, replace=True)
        values.append(metric_f(sample, **metric_f_kwargs))

    # Shapiro-Wilk test
    res = shapiro(values)

    if do_plot_distribution:
        plt.hist(values, bins=100)
        plt.title(f"Distribution for {metric_f.__name__}. Normality p-value: {res.pvalue}")
        plt.show()

    return res.pvalue


def generate_data(
    size: int,
    distribution_type: Literal["cont", "disc"],
    add_index: bool = True,
    index: Union[pd.Index, pd.MultiIndex] = None,
) -> pd.Series:
    """
    Generates data for continuous variables test. Regression technique require special multi-index [entity, dt].
    :param size: sample size for generation
    :param distribution_type: type of variable for generation
    * cont - continuous variable from normal distribution with mean=0 and std=3
    * disc - discrete variable from binomial distribution with p=0.2
    :param add_index: whether to add multiindex or not
    :param index: index for generated sample if add_index is true.
    If add_index is false then multiindex will be generated
    :return:  sample
    """

    if distribution_type == "cont":
        variable = np.random.normal(0, 3, size=size)
    elif distribution_type == "disc":
        variable = np.random.choice([0, 1], size=size, p=[0.8, 0.2])
    else:
        raise ValueError("type should take one of next values: ['cont', 'disc']")

    if add_index:
        if index is None:
            # Generate random index
            entity_index = np.random.choice([1, 2, 3], size=size)
            time_index = np.arange(0, size, 1)
            result_sr = pd.Series(variable, index=[entity_index, time_index])
        else:
            # Add existing index
            result_sr = pd.Series(variable, index=index)
    else:
        # Without index
        result_sr = pd.Series(variable)
    return result_sr


class BaseSimulationClass:
    """
    Virtual class for AA and AB tests simulation
    """

    def __init__(
        self,
        alternative: Literal["less", "greater", "two-sided"],
        stattests_list: List[str],
        sample_size: int,
        experiments_num: int,
        mde: float,
        alpha_level: float = 0.05,
        power: float = 0.8,
    ):
        self.sample_size = sample_size
        self.alternative = alternative
        self.mde = mde
        self.power = power
        self.alpha_level = alpha_level
        self.stattests_list = stattests_list
        self.experiments_num = experiments_num
        self.info = {}
        self.stattests_func_map = {}

    def plot_p_values(self):
        """
        Plot p-values distribution for each test
        :return: None
        """
        if len(self.info) == 0:
            return

        # AB
        x_axis = np.linspace(0, 1, 1000)
        for test, test_info in self.info.items():
            ab_pvalues = np.array(test_info["ab_pvalues"])
            y_axis = [np.mean(ab_pvalues < x) for x in x_axis]
            plt.plot(x_axis, y_axis, label=test)

        plt.plot([self.alpha_level, self.alpha_level], [0, 1], "--k", alpha=0.8)
        plt.plot([0, 1], [self.power, self.power], "--k", alpha=0.8)
        plt.title("P-Value Distribution for AB Simulation", size=12)
        plt.xlabel("p-value", size=10)
        plt.legend(fontsize=10)
        plt.grid()
        plt.show()

        # AA
        for test, test_info in self.info.items():
            aa_pvalues = np.array(test_info["aa_pvalues"])
            y_axis = [np.mean(aa_pvalues < x) for x in x_axis]
            plt.plot(x_axis, y_axis, label=test)
        plt.plot([0, 1], [0, 1], "--k", alpha=0.8)
        plt.title("P-Value Distribution for AA Simulation", size=12)
        plt.xlabel("p-value", size=10)
        plt.legend(fontsize=10)
        plt.grid()
        plt.show()

    def print_results(self):
        """
        Print simulation results for each test (alpha and power + confidence intervals)
        :return: None
        """
        for test_name, test_info in self.info.items():
            a, p = test_info["alpha"], test_info["power"]
            aci1, aci2 = round(test_info["alpha_ci"][0], 4), round(test_info["alpha_ci"][1], 4)
            pci1, pci2 = round(test_info["power_ci"][0], 4), round(test_info["power_ci"][1], 4)

            if (aci1 > self.alpha_level) or (self.power > pci2):
                # fail
                print("\033[91m" + f"'{test_name}'; alpha={a} [{aci1}; {aci2}], power={p} [{pci1}; {pci2}]" + "\033[0m")
            else:
                # success
                print("\033[92m" + f"'{test_name}'; alpha={a} [{aci1}; {aci2}], power={p} [{pci1}; {pci2}]" + "\033[0m")

    def run(self):
        """
        Simulate all tests from 'self.stattests_list' by given data and save information to 'info' dictionary
        :return:
        """
        self.info = {}
        for stattest in self.stattests_list:
            self.simulate_test_by_name(stattest)
        return self.info

    def simulate_test_by_name(self, test_name: str):
        """
        Simulate AA and AB test and save results to 'info' dictionary
        :param test_name: name of test for simulation (ttest | cuped_ttest | regression_test | did_regression_test
                                                        | additional_vars_regression_test)
        :return: None
        """

        assert test_name in self.stattests_func_map, f"Given test_name {test_name} not found"
        stattest_func = self.stattests_func_map[test_name]

        test_success_no_effect_cnt = 0
        test_pvalues_no_effect = []
        test_success_effect_cnt = 0
        test_pvalues_effect = []

        for _ in tqdm(range(self.experiments_num), desc=f"Simulation test '{test_name}'"):
            p_value = stattest_func(mde=0)
            test_pvalues_no_effect.append(p_value)
            if p_value < self.alpha_level:
                test_success_no_effect_cnt += 1

            p_value = stattest_func(mde=self.mde)
            test_pvalues_effect.append(p_value)
            if p_value < self.alpha_level:
                test_success_effect_cnt += 1

        alpha = test_success_no_effect_cnt / self.experiments_num
        power = test_success_effect_cnt / self.experiments_num

        alpha_ci = estimate_ci_binomial(alpha, self.experiments_num, alpha=0.05)
        power_ci = estimate_ci_binomial(power, self.experiments_num, alpha=0.05)

        if test_name in self.info:
            del self.info[test_name]
        self.info[test_name] = {
            "alpha": alpha,
            "alpha_ci": alpha_ci,
            "power": power,
            "power_ci": power_ci,
            "aa_pvalues": test_pvalues_no_effect,
            "ab_pvalues": test_pvalues_effect,
        }



import unittest

import numpy as np
import pandas as pd

from abtoolkit.continuous.utils import estimate_mde_by_sample_size
from abtoolkit.continuous.utils import estimate_sample_size_by_mde


class TestSampleSizeEstimation(unittest.TestCase):
    def test_calculate_sample_size_by_mde(self):
        sample_size = estimate_sample_size_by_mde(3, 0.05, 0.8, 2, "two-sided")
        self.assertTrue(isinstance(sample_size, int), f"Estimated sample size has wrong type: {sample_size}")
        self.assertTrue(sample_size > 0, f"Estimated sample size has negative value: {sample_size}")

    def test_calculate_mde_by_sample_size(self):
        mde = estimate_mde_by_sample_size(3, 0.05, 0.8, 36, "two-sided")
        self.assertTrue(isinstance(mde, float), f"Estimated MDE has wrong type: {mde}")



"""
Utils functions for continuous variables
"""

from typing import Tuple, Literal

import numpy as np
from scipy import stats


def estimate_confidence_interval(
    mean: float,
    std: float,
    sample_size: int,
    alpha: float,
    power: float,
    alternative: Literal["less", "greater", "two-sided"],
) -> Tuple[float, float]:
    """
    Confidence interval estimation
    :param mean: sample average value
    :param std: sample standard deviation
    :param sample_size: number of samples
    :param alpha: alpha-level
    :param power: power
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: low confidence interval value, high confidence interval value
    """
    if alternative == "two-sided":
        alpha = alpha / 2

    z = stats.norm.ppf(q=1 - alpha) + stats.norm.ppf(q=power)
    se = z * std / np.sqrt(sample_size)

    return mean - se, mean + se


def estimate_sample_size_by_mde(
    std: float,
    alpha: float,
    power: float,
    mde: float,
    alternative: Literal["less", "greater", "two-sided"],
) -> int:
    """
    Calculate sample size need for significant ttest using std, alpha, power, mde
    :param std: tested variable std
    :param alpha: level of significance of A/B test
    :param power: probability of observing a statistically significant result at level alpha
    if a true effect of a certain magnitude is present
    :param mde: minimal detectable effect, the difference in results to detect
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: sample size needed for each group
    """

    if alternative == "two-sided":
        alpha = alpha / 2

    z = stats.norm.ppf(q=1 - alpha) + stats.norm.ppf(q=power)
    size = 2 * (std * z / mde) ** 2

    return round(size) + 1


def estimate_mde_by_sample_size(
    std: float,
    alpha: float,
    power: float,
    sample_size: int,
    alternative: Literal["less", "greater", "two-sided"],
):
    """
    Calculate sample size need for significant ttest using std, alpha, power, mde
    :param std: tested variable std
    :param alpha: level of significance of A/B test
    :param power: probability of observing a statistically significant result at level alpha
    if a true effect of a certain magnitude is present
    :param sample_size: sample size needed for each group
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : means are equal;
    * 'less': the mean of the control sample is less than the mean of the test sample;
    * 'greater': the mean of the control sample is greater than the mean of the test sample;
    :return: minimal detectable effect, the difference in results to detect
    """
    if alternative == "two-sided":
        alpha = alpha / 2

    z = stats.norm.ppf(q=1 - alpha) + stats.norm.ppf(q=power)
    mde = std * z / np.sqrt(sample_size / 2)

    return mde



from abtoolkit.continuous.simulation import StatTestsSimulation
from abtoolkit.continuous.utils import estimate_sample_size_by_mde
from abtoolkit.utils import generate_data


if __name__ == '__main__':
    # Fix global params
    mde = 2
    alpha_level = 0.05
    power = 0.8
    examples_num = 2000  # Number of examples in test and control group
    experiments_num = 200  # Number of experiments to run for each stattest
    alternative = 'two-sided'

    # Generate variable
    variable = generate_data(examples_num, distribution_type="cont")
    # Generate previous value of variable (will be used to reduce variance) and speedup tests
    previous_value = generate_data(examples_num, distribution_type="cont", index=variable.index).rename("prev")

    # Estimate sample_size need for test
    sample_size = estimate_sample_size_by_mde(
        std=variable.std(),
        alpha=alpha_level, power=power, mde=mde, alternative=alternative)
    print(f"Minimum sample size for each group is {sample_size}")

    sim = StatTestsSimulation(
        variable,
        stattests_list=["ttest", "diff_ttest", "regression_test", "cuped_ttest", "did_regression_test",
                        "additional_vars_regression_test"],
        alternative=alternative,
        experiments_num=experiments_num,  # Run each stattest 10 times
        sample_size=sample_size,  # Take 50 samples from variables
        mde=mde,  # Trying to detect this effect (very big for our simulated data)
        alpha_level=alpha_level,  # Fix alpha level on 5%

        previous_values=previous_value,
        cuped_covariant=previous_value,
        additional_vars=[previous_value],
    )
    info = sim.run()  # Get dictionary with information about tests
    sim.print_results()  # Print results of simulation
    sim.plot_p_values()  # Plot p-values distribution



"""
Stat tests for discrete variable analysis
"""

from typing import Literal

import numpy as np
from scipy import stats
from abtoolkit.discrete.utils import compare_beta_distributions


def conversion_ztest(
    control_count: int,
    control_objects_num: int,
    test_count: int,
    test_objects_num: int,
    alternative: Literal["less", "greater"],
) -> float:
    """
    Conversion z-test
    :param control_count: number of positive samples in control group
    :param control_objects_num: number of all samples in control group
    :param test_count: number of positive samples in test group
    :param test_objects_num: number of all samples in test group
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'two-sided' : conversions are equal;
    * 'less': the conversion of the control sample is less than the mean of the test sample;
    * 'greater': the conversion of the control sample is greater than the mean of the test sample;
    :return: p-value
    """

    diff = control_count / control_objects_num - test_count / test_objects_num
    p_pooled = (control_count + test_count) / (control_objects_num + test_objects_num)
    std_diff = np.sqrt(p_pooled * (1 - p_pooled) * (1 / control_objects_num + 1 / test_objects_num))
    z_stat = diff / std_diff

    if alternative == "less":
        p_value = stats.norm.cdf(z_stat)
    elif alternative == "greater":
        p_value = stats.norm.sf(z_stat)
    elif alternative == "two-sided":
        p_value = stats.norm.sf(np.abs(z_stat)) * 2
    else:
        raise ValueError("alternative must be 'less', 'greater' or 'two-sided'")

    return p_value


def bayesian_test(
    control_count: int,
    control_objects_num: int,
    test_count: int,
    test_objects_num: int,
    alternative: Literal["less", "greater"],
    prior_positives_count: int = 1,
    prior_negatives_count: int = 1,
) -> float:
    """
    Bayesian Test
    :param control_count: posterior number of positive samples in control group
    :param control_objects_num: posterior number of all samples in control group
    :param test_count: posterior number of positive samples in test group
    :param test_objects_num: posterior number of all samples in test group
    :param alternative: alternative hypothesis ("less", "greater" or "two-sided").
    * 'less': the conversion of the control sample is less than the convertion in the test sample;
    * 'greater': the conversion of the control sample is greater than the conversion in the test sample;
    :param prior_positives_count: prior number of positive samples (alpha in Beta distribution)
    :param prior_negatives_count: prior number of negative samples (beta in Beta distribution)
    :return: probability that difference between distributions not exists (probability of null hypothesis)
    """

    # calculate posterior params for distributions
    alpha_1, beta_1 = test_count + prior_positives_count, test_objects_num - test_count + prior_negatives_count
    alpha_2, beta_2 = control_count + prior_positives_count, control_objects_num - control_count + prior_negatives_count

    if alternative == "less":
        return compare_beta_distributions(alpha_1, beta_1, alpha_2, beta_2)

    return compare_beta_distributions(alpha_2, beta_2, alpha_1, beta_1)



import unittest

import numpy as np
from abtoolkit.discrete.stattests import conversion_ztest
from abtoolkit.discrete.stattests import bayesian_test
from abtoolkit.utils import generate_data


class TestStatTests(unittest.TestCase):
    def test_ztest(self):
        test_sr = generate_data(100, distribution_type="disc")
        control_sr = generate_data(100, distribution_type="disc")
        p_value = conversion_ztest(
            control_sr.sum(),
            len(control_sr),
            test_sr.sum(),
            len(test_sr),
            "less"
        )
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")

    def test_bayesian_test(self):
        test_sr = generate_data(100, distribution_type="disc")
        control_sr = generate_data(100, distribution_type="disc")
        p_value = bayesian_test(
            control_sr.sum(),
            len(control_sr),
            test_sr.sum(),
            len(test_sr),
            'less'
        )
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")

    def test_bayesian_test_prior(self):
        test_sr = generate_data(100, distribution_type="disc")
        control_sr = generate_data(100, distribution_type="disc")
        p_value = bayesian_test(
            control_sr.sum(),
            len(control_sr),
            test_sr.sum(),
            len(test_sr),
            'less',
            prior_positives_count=10,
            prior_negatives_count=10,
        )
        self.assertTrue(0 <= p_value <= 1, f"Wrong value for p-value: {p_value}")



import numpy as np

from abtoolkit.discrete.simulation import StatTestsSimulation
from abtoolkit.discrete.utils import estimate_sample_size_by_mde
from abtoolkit.utils import generate_data


if __name__ == '__main__':
    # Fix global params
    mde = 0.05
    alpha_level = 0.05
    power = 0.8
    examples_num = 2000  # Number of examples in test and control group
    experiments_num = 1000  # Number of experiments to run for each stattest
    alternative = 'less'

    # Generate variable
    variable = generate_data(examples_num, distribution_type="disc")

    proba = variable.sum() / variable.count()

    # Estimate sample_size need for test
    sample_size = estimate_sample_size_by_mde(p=proba, alpha=alpha_level, power=power, mde=mde, alternative=alternative)
    print(f"Minimum sample size for each group is {sample_size}")

    sim = StatTestsSimulation(
        count=variable.sum(),
        objects_num=variable.count(),
        stattests_list=["conversion_ztest", "bayesian_test"],
        alternative=alternative,
        experiments_num=experiments_num,  # Run each stattest 10 times
        sample_size=sample_size,  # Take 50 samples from variables
        mde=mde,  # Trying to detect this effect (very big for our simulated data)
        alpha_level=alpha_level,  # Fix alpha level on 5%
        power=power,
        bayesian_prior_positives=1,
        bayesian_prior_negatives=1,
    )
    info = sim.run()  # Get dictionary with information about tests
    sim.print_results()  # Print results of simulation
    sim.plot_p_values()  # Plot p-values distribution
